#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#include <cerrno>
#include <cstring>
#include <fstream>
#include <iostream>
#include <memory>
#include <sstream>
#include <string>
#include <vector>

#include <boost/program_options/options_description.hpp>
#include <boost/program_options/parsers.hpp>
#include <boost/program_options/positional_options.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/scoped_array.hpp>

const unsigned	TERM_COLS = 80;
char		*prog;

class Progress;

static void	bestzip(const std::string &);
static blksize_t
		file_block_size(const std::string &);
static bool	file_remove(const std::string &);
static ssize_t	file_size(const std::string &);
static bool	pipe_file(const std::string &, const std::string &,
		    char *const *, Progress &, blksize_t);

class Pipe {
public:
	struct error {};

	Pipe()
	{
		if (pipe(fds) == -1)
			throw error();
	}

	~Pipe()
	{
		if (fds[0] > 0) ::close(fds[0]);
		if (fds[1] > 0) ::close(fds[1]);
	}

	int close_in()
	{
		int retval = 0;
		if (fds[1] > 0) {
			retval = ::close(fds[1]);
			fds[1] = -1;
		}
		return retval;
	}

	int close_out()
	{
		int retval = 0;
		if (fds[0] > 0) {
			retval = ::close(fds[0]);
			fds[0] = -1;
		}
		return retval;
	}

	int close()
	{
		int retval = 0;
		if (close_in() == -1)
			retval = -1;
		if (close_out() == -1)
			retval = -1;
		return retval;
	}

	int in() const
	{ return fds[1]; }

	int out() const
	{ return fds[0]; }

private:
	int fds[2];
};

class Progress {
public:
	Progress(std::ostream &out, const std::string &label) :
	    out(out), label(label)
	{
		// format of a line (three chars for percentage):
		// label  " |"  .{cols}  "| "  percentage "%"
		if (TERM_COLS < label.length() + 8) {
			cols = 0;
		} else {
			cols = TERM_COLS - label.length() - 8;
		}
	}

	void update(float proportion)
	{
		out << '\r' << label << ' ';

		if (cols) {
			unsigned char	stars;
			unsigned char	i;
			stars = 
			    static_cast<unsigned char>(cols * proportion);
			out << '|';
			for (i = 0; i < stars; i++)
				out << '*';
			for (; i < cols; i++)
				out << ' ';
			out << "| ";
			out.width(3);
		}
		// else, nothing fancy

		out << static_cast<int>(proportion * 100) << '%';
	}

	void finished(size_t size)
	{
		std::ostringstream strout;
		int blanks;
		strout << label << ": " << size << " bytes";
		blanks = TERM_COLS - strout.str().length();

		out << '\r' << strout.str();
		while (blanks--)
			out.put(' ');
		out.put('\n');
	}

private:
	std::ostream &out;
	const std::string &label;
	unsigned char cols;
};

static void
bestzip(const std::string &path)
{
	static char args[3][5];
	static char *gzip[3];
	static char *lzma[3];
	static bool init = false;

	if (!init) {
		strcpy(args[0], "gzip");
		strcpy(args[1], "lzma");
		strcpy(args[2], "-9");

		gzip[0] = args[0];
		gzip[1] = args[2];
		gzip[2] = 0;

		lzma[0] = args[1];
		lzma[1] = args[2];
		lzma[2] = 0;

		init = true;
	}

	blksize_t	block_size;
	ssize_t		size_gzip;
	ssize_t		size_lzma;

	std::cout << path << ":\n";

	if (!(block_size = file_block_size(path))) {
		std::cerr << "skipping " << path << '\n';
		return;
	}

	std::string path_gzip(path);
	path_gzip += ".gz";
	Progress prog_gzip(std::cout, path_gzip);
	if (!pipe_file(path, path_gzip, gzip, prog_gzip, block_size)) {
		std::cerr << "skipping " << path << '\n';
		return;
	}

	if ((size_gzip = file_size(path_gzip)) == -1) {
		std::cout << '\n';
		std::cerr << "skipping " << path << '\n';
		return;
	}
	prog_gzip.finished(size_gzip);

	std::string path_lzma(path);
	path_lzma += ".lzma";
	Progress prog_lzma(std::cout, path_lzma);
	if (!pipe_file(path, path_lzma, lzma, prog_lzma, block_size)) {
		std::cerr << "skipping " << path << '\n';
		return;
	}

	if ((size_lzma = file_size(path_lzma)) == -1) {
		std::cout << '\n';
		std::cerr << "skipping " << path << '\n';
		return;
	}
	prog_lzma.finished(size_lzma);

	std::string path_larger;
	if (size_gzip > size_lzma)
		path_larger = path_gzip;
	else
		path_larger = path_lzma;

	file_remove(path_larger);
}

static blksize_t
file_block_size(const std::string &path)
{
	struct stat st;
	if (stat(path.c_str(), &st) == -1) {
		std::cerr << prog << ": could not stat " << path << ": "
		    << strerror(errno) << '\n';
		return 0;
	}
	return st.st_blksize;
}

static bool
file_remove(const std::string &path)
{
	if (unlink(path.c_str()) == -1) {
		std::cerr << prog << ": could not remove " << path << ": "
		    << strerror(errno) << '\n';
		return false;
	}
	return true;
}

static ssize_t
file_size(const std::string &path)
{
	struct stat st;
	if (stat(path.c_str(), &st) == -1) {
		std::cerr << prog << ": could not stat " << path << ": "
		    << strerror(errno) << '\n';
		return -1;
	}
	return static_cast<ssize_t>(st.st_size);
}

static bool
pipe_file(const std::string &path_in, const std::string &path_out,
    char *const *args, Progress &progress, blksize_t block_size)
{
	std::auto_ptr<Pipe> pipe;
	int	fd_out;
	pid_t	pid;

	try {
		pipe.reset(new Pipe);
	} catch (Pipe::error) {
		std::cerr << prog << ": creating pipe: " << strerror(errno)
		    << '\n';
		return false;
	}

	if ((fd_out = open(path_out.c_str(),
	    O_CREAT | O_TRUNC | O_WRONLY,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) == -1) {
		std::cerr << prog << ": creating " << path_out << ": "
		    << strerror(errno) << '\n';
		return false;
	}

	if ((pid = fork()) == -1) {
		std::cerr << prog << ": forking child: " << strerror(errno)
		    << '\n';
		close(fd_out);
		return false;
	} else if (!pid) {
		pipe->close_in();

		dup2(pipe->out(), STDIN_FILENO);
		dup2(fd_out, STDOUT_FILENO);

		execvp(*args, args);
		exit(0xff);
	}

	pipe->close_out();
	close(fd_out);
	fd_out = -1;

	std::ifstream in(path_in.c_str(),
	    std::ios_base::binary | std::ios_base::in);

	// file already stat'd, so this should be fine
	ssize_t size = file_size(path_in);

	boost::scoped_array<char> buf(new char[block_size]);
	size_t total = 0;
	size_t bytes;
	size_t offset;
	while (in) {
		in.read(buf.get(), block_size);
		bytes = in.gcount();
		total += bytes;
		offset = 0;
		while (bytes) {
			ssize_t retval = write(pipe->in(),
			    buf.get() + offset, bytes);
			if (retval == -1 && errno == EINTR) continue;
			if (retval == -1) {
				std::cerr << prog << ": pipe write error: "
				    << strerror(errno)
				    << "\nterminating child process\n";
				kill(pid, SIGTERM);
				file_remove(path_out);
				return false;
			}

			bytes -= retval;
			offset += retval;
		}

		progress.update(static_cast<float>(total) / size);
	}

	pipe->close();
	in.close();

	int status;
	do {
		status = 0;
		pid_t retval = waitpid(pid, &status, 0);
		if (retval == -1 && errno == EINTR) continue;
		if (retval == -1) {
			std::cerr << prog << ": waitpid failed: "
			    << strerror(errno) << '\n';
			return false;
		}
	} while (false);

	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) == 0xff) {
			std::cerr << prog << ": failed to exec " << *args
			    << '\n';
			file_remove(path_out);
			return false;
		} else if (WEXITSTATUS(status)) {
			std::cerr << prog << ": " << *args
			    << " failed (error code: " << WEXITSTATUS(status)
			    << ")\n";
			file_remove(path_out);
			return false;
		}
	} else {
		std::cerr << prog << ": " << *args << " was terminated\n";
		file_remove(path_out);
		return false;
	}

	return true;
}

int
main(int argc, char **argv)
{
	using namespace boost::program_options;
	using std::cout;

	prog = *argv;

	// describe options
	options_description options;
	options.add_options()
	    ("help", "Show this message.")
	    ;

	// describe positional
	options_description hidden;
	hidden.add_options()
	    ("input", value<std::vector<std::string> >(), "")
	    ;
	positional_options_description positional;
	positional.add("input", -1);

	// combine general options and hidden options to pass to parser
	options_description combined;
	combined.add(options).add(hidden);

	// set up parser
	command_line_parser parser(argc, argv);
	parser.options(combined);
	parser.positional(positional);

	// parse
	variables_map vars;
	store(parser.run(), vars);

	if (vars.count("help") || !vars.count("input")) {
		cout << "usage: " << prog << " [OPTION...] FILE...\n"
		    << options;
		return 1;
	}

	std::vector<std::string> files =
	    vars["input"].as<std::vector<std::string> >();

	typedef std::vector<std::string>::iterator Iter;
	for (Iter iter = files.begin(); iter != files.end(); ++iter) {
		bestzip(*iter);
	}

	return 0;
}
